#!/bin/bash

AUTHOR="Janne Hellas"
VERSION="0.4"
LICENSE="Public Domain or equivalent."
LINFO="This software is provided \"as-is\" without any warranties or liabilities."

SCRIPTDIR=$(dirname "${BASH_SOURCE}")

# Test environment
BC=`which bc`
if [[ "${BC}" == "" ]]; then
   echo "$0: Environmental failure. Install bc." 1>&2
   exit 1
fi

if [[ $# -eq 0 ]]; then
   echo "FF3/6 SNES to GBA save converter v$VERSION"
   echo ""
   echo "Usage: ./ff6gba_snessave_import.sh [SNES Save file] [Donor GBA save file]"
   echo ""
   echo "The utility will output a file called 'data_import.sav' from the input files."
   echo "Donor file's slot 1 data blocks are replaced by the migrated data from the input SNES save file's slot 1."
   echo "Bestiary and other slots are augmented from the donor file."
   echo "If no donor is specified, the other slots and bestiary are created empty."
   echo "Nothing is written to the input files."
   echo ""
   echo "License: $LICENSE"
   echo "$LINFO"
   exit 0
fi

DONOR="/tmp/empty_gba$$.sav"
# handle non-option arguments
if [[ $# -gt 2 ]]; then
    echo "$0: Max two arguments are required." 1>&2
    exit 1
fi

if [[ ! -f "$1" ]]; then
   echo "$0: SNES save file not found." 1>&2
   exit 1
fi

# Detect ff3 snes
FF3SNES=`"${SCRIPTDIR}/util/ff6detect.sh" "$1"`
if [ "${FF3SNES}" != "ff3snes" ]; then
   echo "$0: SNES save file not detected in input file." 1>&2
   exit 2
fi

# Static parameters
SLOT=1
OUTFILE="data_import.sav"

if [[ "$1" == "$OUTFILE" ]] || [[ "$2" == "$OUTFILE" ]]; then
   echo "$0: Input file cannot be named the same as output." 1>&2
   exit 3
fi

if [[ -f "$OUTFILE" ]]; then
    echo "Output file '$OUTFILE' exists. Overwrite? (y/n)"
    read OVERWRITE
    if [ $OVERWRITE == "yes" ] || [ $OVERWRITE == "y" ]; then
       echo "Will overwrite."
       rm "$OUTFILE"
    else
       echo "Nothing to do. Exiting."
       exit 0
    fi
fi

if [[ ! -f "$2" ]]; then
   echo "Donor file not found. Using an empty one."
   dd if=/dev/zero bs=1 count=23552 status=none > "$DONOR"
   dd if=/dev/zero bs=1 count=41984 status=none | tr "\000" "\377" >> "$DONOR"
else
   DONOR="$2"
   # Detect donor is ff6 advance
   FF6ADV=`"${SCRIPTDIR}/util/ff6detect.sh" "$DONOR"`
   if [ "${FF6ADV}" != "ff6advance" ]; then
      echo "$0: Donor file not detected as ff6advance." 1>&2
      exit 2
   fi
fi

echo "Internal checks completed. Starting import."

# Data slot
echo "Importing data slot"

# Item data
ITEM_OFFSET=617
ITEMADV_OFFSET=8208
# Item data slot size for items in advance version
ITEMSLOTADV_SIZE_OFFSET=1024

# Inventory size, bytes
ITEMS_SIZE=256
ITEMSADV_SIZE=288

ITEMHEX=`xxd -p -l $ITEMS_SIZE -s "$ITEM_OFFSET" -u "$1" | tr -d " \n"`
NBHEX=`xxd -p -l $ITEMS_SIZE -s "$(( $ITEM_OFFSET + $ITEMS_SIZE ))" -u "$1" | tr -d " \n"`

#echo $ITEMHEX
#echo $NBHEX

ESPERHEX=`xxd -p -l 4 -s 1129 -u "$1" | tr -d " \n"`
ACTIVEGROUPE=`xxd -p -l 1 -s 1133 -u "$1" | tr -d " \n"`
MORPHBAR=`xxd -p -l 1 -s 1782 -u "$1" | tr -d " \n"`
BUSHIDO=`xxd -p -l 1 -s 1783 -u "$1" | tr -d " \n"`

#echo $ESPERHEX
#echo $ACTIVEGROUPE

head -c $ITEM_OFFSET "$1" >> "$OUTFILE"

# Magic, 54 bytes per character
MAGIC_OFFSET=1134
# 54 spells
MAGICDATASLOT_SIZE=54
# 12 slots for 12 characters
MAGICSLOTS=12

# For Advance version
MAGICADV_OFFSET=617
EXTRA_OFFSET=7
MAGICDATASLOTADV_SIZE=57

MAGICHEX=`xxd -p -l $(( $MAGICDATASLOT_SIZE * $MAGICSLOTS )) -s $MAGIC_OFFSET -u "$1" | tr -d " \n"`

#echo $MAGICHEX

# Patch Magic data

for (( i=0; i<${MAGICSLOTS}; i++ ))
do
   echo ${MAGICHEX:$(( $i * $MAGICDATASLOT_SIZE * 2 )):$(( $MAGICDATASLOT_SIZE * 2 ))} | xxd -p -r >> "$OUTFILE"
   # Zero GBA specific spells & extra space
   dd if=/dev/zero bs=1 count=10 status=none >> "$OUTFILE"
done

# More zeros
dd if=/dev/zero bs=1 count=396 status=none >> "$OUTFILE"

# Write active group number
echo ${ACTIVEGROUPE} | xxd -p -r >> "$OUTFILE"

# Write morphbar
echo "${MORPHBAR}" | xxd -p -r >> "$OUTFILE"

# Write bushido
echo "${BUSHIDO}" | xxd -p -r >> "$OUTFILE"

# Bushido name marks
echo "D3AA9EEF90FFD3AA9EEF9AFFD3AA9EEFA0FFD3AA9EEFD5FFD3AA9EEFE9FFD3AA9EEF9DFFD3AA9EEFECFFD3AA9EEFC2FF" | xxd -p -r >> "$OUTFILE"

# Copy until empty 0x55 blocks and zero them
tail -c +1833 "$1" | head -c 245 >> "$OUTFILE"
dd if=/dev/zero bs=1 count=35 status=none >> "$OUTFILE"

tail -c +2113 "$1" | head -c 48 >> "$OUTFILE"
dd if=/dev/zero bs=1 count=16 status=none >> "$OUTFILE"

tail -c +2177 "$1" | head -c 322 >> "$OUTFILE"

tail -c +2499 "$1" | head -c 16 | sed 's/\x55\x55/\x00\x00/g' >> "$OUTFILE"

# Parent facing direction
tail -c +2515 "$1" | head -c 1 >> "$OUTFILE"

# Saved XY
for (( i=0; i<32; i=i+2 ))
do
   tail -c +$(( 2516 + $i * 2 )) "$1" | head -c 2 | sed 's/\x55\x55/\x00\x00/g' >> "$OUTFILE"
done

# Z-levels
tail -c +2548 "$1" | head -c 4 >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=9 status=none >> "$OUTFILE"

# Reset save counter, just in case
printf "$(printf '\\x%02X' 0x00)" | dd of="$OUTFILE" bs=1 seek=1991 count=1 conv=notrunc status=none
printf "$(printf '\\x%02X' 0x00)" | dd of="$OUTFILE" bs=1 seek=1992 count=1 conv=notrunc status=none

# Other slots
tail -c +2561 "$DONOR" | head -c 5376 >> "$OUTFILE"

# FF6 Advance marks + save mark
echo "46494E414C2046414E54415359205649" | xxd -p -r >> "$OUTFILE"
echo "2020202020414456414E43452020202001" | xxd -p -r >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=207 status=none >> "$OUTFILE"

# Slot marks
echo "01000000000000001BE41BE41BE41BE4" | xxd -p -r >> "$OUTFILE"


# Data extension

# Espers
echo $ESPERHEX | xxd -p -r >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=12 status=none >> "$OUTFILE"

# Items
echo "$ITEMHEX" | xxd -p -r >> "$OUTFILE"

# Empty GBA specific inventory
echo "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF" | xxd -p -r >> "$OUTFILE"

# Item counts
echo "$NBHEX" | xxd -p -r >> "$OUTFILE"

# Empty GBA specific counts + GBA eventbits
dd if=/dev/zero bs=1 count=48 status=none >> "$OUTFILE"

# Zeros
dd if=/dev/zero bs=1 count=412 status=none >> "$OUTFILE"

# Extension slot mark + space for checksum
echo "474D0000" | xxd -p -r >> "$OUTFILE"

# Test whether to add Stone Table to inventory
TEST=$(( 0x${ESPERHEX:2:2} & 0x80 ))
if [[ $TEST -eq 128 ]]; then
   echo "Adding Stone Tablet to inventory."
   RAREI=`xxd -p -l 1 -s 2236 -u "$1"`
   NEWVALUE=$(( 0x$RAREI | 0x10 ))
   NEWVALUE2=$(echo "obase=16; ibase=10; $NEWVALUE" | bc)
   #echo "Replace debug: ${ESPERHEX:2:2}, $RAREI, $NEWVALUE, $NEWVALUE2"
   printf "$(printf '\\x%02X' 0x$NEWVALUE2)" | dd of="$OUTFILE" bs=1 seek=2236 count=1 conv=notrunc status=none
fi

# Rest of the data (Includes other slots + bestiaries + quicksave slot)
tail -c +9217 "$DONOR" | head -c 56320 >> "$OUTFILE"

# Remove possibly created empty donor file.
if [[ -f "/tmp/empty_gba$$.sav" ]]; then
   rm "/tmp/empty_gba$$.sav"
fi

# Names
for (( i=0; i<16; i++ ))
do
   NAMEHEX=`xxd -p -l 6 -s $(( 2 + $i * 37 )) -u "$1" | tr -d " \n"`
   NAME=`"${SCRIPTDIR}/util/ff6name.sh" "$NAMEHEX"`
   echo "Import name: $NAME"

   INAMEHEX=""
   for (( j=0; j<6; j++ ))
   do
      CHAR=${NAME:$j:1}
      ICHARHEX=`grep ";$CHAR;" "$SCRIPTDIR/util/crossmap" | awk -F ';' '{print $3}'`
      INAMEHEX="$INAMEHEX$ICHARHEX"
   done
   #echo "Debug import hex: $INAMEHEX"

   OF=0
   for (( k=0; k<${#INAMEHEX}; k=k+2 ))
   do
      printf "$(printf '\\x%02X' 0x${INAMEHEX:$k:2})" | dd of="$OUTFILE" bs=1 seek=$(( 2 + $OF + $i * 37 )) count=1 conv=notrunc status=none
      OF=$(( $OF + 1 ))
   done
done

echo "Import completed; '$OUTFILE' created."
echo "Running checksum updater."

"$SCRIPTDIR/util/ff6gba_checksumupdater.sh" "$OUTFILE"
